from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from typing import Any, List, Dict

# Backward compatibility: try both import methods
try:
    # New way (package structure)
    from ..config import POSTGRES_USER, POSTGRES_PASSWORD, DB_HOST, DB_PORT, POSTGRES_DB
except ImportError:
    # Old way (direct import for backward compatibility)
    try:
        from config import POSTGRES_USER, POSTGRES_PASSWORD, DB_HOST, DB_PORT, POSTGRES_DB
    except ImportError:
        # Fallback to environment variables
        import os
        POSTGRES_USER = os.getenv('POSTGRES_USER', 'cellframe')
        POSTGRES_PASSWORD = os.getenv('POSTGRES_PASSWORD', '')
        DB_HOST = os.getenv('POSTGRES_HOST', 'localhost')
        DB_PORT = os.getenv('POSTGRES_PORT', '5432')
        POSTGRES_DB = os.getenv('POSTGRES_DB', 'cellframe')

DATABASE_URI = f'postgresql://{POSTGRES_USER}:{POSTGRES_PASSWORD}@{DB_HOST}:{DB_PORT}/{POSTGRES_DB}'

engine = create_engine(DATABASE_URI)
session = sessionmaker(autocommit=False, autoflush=False, bind=engine)

# Create Base using SQLAlchemy v1.4 syntax
Base = declarative_base()

# Add methods to Base
@classmethod 
def model_lookup_by_table_name(cls, table_name):
    registry_instance = getattr(cls, "registry")
    for mapper_ in registry_instance.mappers:
        model = mapper_.class_
        if not hasattr(model, "__tablename__"):
            model_class_name = model.__table__.name
        else:
            model_class_name = model.__tablename__
        if model_class_name == table_name:
            return model

@classmethod
def schema(cls) -> str:
    """Return name of database schema the model refers to."""

    _schema = cls.__mapper__.selectable.schema
    if _schema is None:
        raise ValueError("Cannot identify model schema")
    return _schema

@classmethod
def table_name(cls) -> str:
    """Return name of the table the model refers to."""

    return cls.__tablename__

@classmethod
def fields(cls) -> List[str]:
    """Return list of model field names."""

    return cls.__mapper__.selectable.c.keys()

def to_dict(self) -> Dict[str, Any]:
    """Convert model instance to a dictionary."""

    _dict: Dict[str, Any] = dict()
    for key in self.__mapper__.c.keys():
        _dict[key] = getattr(self, key)
    return _dict

# Add methods to Base class
Base.model_lookup_by_table_name = model_lookup_by_table_name
Base.schema = schema
Base.table_name = table_name
Base.fields = fields
Base.to_dict = to_dict
