import json
from datetime import datetime
from typing import List, Dict, Optional, Union

from sqlalchemy import Table, Column, UniqueConstraint, CheckConstraint, event, func, case, select
from sqlalchemy import ForeignKey, String, Integer, BigInteger
from sqlalchemy.orm import relationship, column_property, object_session
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.types import ARRAY, JSON, TypeDecorator, Unicode, DateTime

from . import Base
from .._types import DatumTypes, ChainTypes

blocks_datums = Table(
    "blocks_datums",
    Base.metadata,
    Column("block_id", ForeignKey("blocks.id"), primary_key=True),
    Column("datum_id", ForeignKey("datums.id"), primary_key=True),
)

events_datums = Table(
    "events_datums",
    Base.metadata,
    Column("event_id", ForeignKey("events.id"), primary_key=True),
    Column("datum_id", ForeignKey("datums.id"), primary_key=True),
)


class DictionaryType(TypeDecorator):
    impl = Unicode

    def process_bind_param(self, value, dialect):
        if isinstance(value, dict):
            return {key: self.process_bind_param(value, dialect) for key, value in value.items()}
        elif isinstance(value, list):
            return [self.process_bind_param(item, dialect) for item in value]
        if not isinstance(value, str):
            value = str(value)
        return value

    def process_result_value(self, value, dialect):
        return json.loads(value)


class DatumBlockAssociation(Base):
    __table__ = blocks_datums


class DatumEventAssociation(Base):
    __table__ = events_datums


class Net(Base):
    __tablename__ = 'nets'
    
    id = Column(BigInteger, primary_key=True)
    name = Column(String, unique=True)
    
    # Relationships will be defined after all classes
    # wallets = relationship("Wallet", back_populates="net")
    # chains = relationship("Chain", back_populates="net")


class Block(Base):
    __tablename__ = 'blocks'
    
    id = Column(String, primary_key=True)  # hash + chain_id
    version = Column(Integer)
    cell_id = Column(String)
    chainid = Column(String)  # TODO: move to Chain
    hash = Column(String, nullable=False)
    created_at = Column(DateTime)
    signs = Column(JSON)
    meta = Column(DictionaryType)
    
    # Foreign keys
    chain_id = Column(BigInteger, ForeignKey('chains.id'), index=True)
    
    # Relationships will be defined after all classes
    # chain = relationship(back_populates="blocks")
    # datums = relationship(secondary=blocks_datums, back_populates="blocks")
    # block_recs = relationship()
    # datums_id = association_proxy("block_recs", "id", creator=lambda d_id: DatumBlockAssociation(datum_id=d_id))

    __table_args__ = (
        UniqueConstraint('hash', 'chain_id', name='uq_block_hash_and_chain_id'),
    )


# auto set id
@event.listens_for(Block, 'before_insert')
def set_id_before_insert(mapper, connection, target):
    target.id = f"{target.hash}{target.chain_id}"


class Event(Base):
    __tablename__ = 'events'
    
    id = Column(String, primary_key=True)  # hash + chain_id
    version = Column(Integer)
    cell_id = Column(String)
    hash = Column(String, nullable=False)
    created_at = Column(DateTime)
    signs = Column(JSON)
    
    # Foreign keys
    chain_id = Column(BigInteger, ForeignKey('chains.id'), index=True)
    
    # Relationships will be defined after all classes
    # chain = relationship(back_populates="events")
    # datums = relationship(secondary=events_datums, back_populates="events")
    # event_recs = relationship()
    # datums_id = association_proxy("event_recs", "id", creator=lambda d_id: DatumEventAssociation(datum_id=d_id))

    # only events fields:
    round_id = Column(Integer)
    hashes = Column(ARRAY(String))

    __table_args__ = (
        UniqueConstraint('hash', 'chain_id', name='uq_event_hash_and_chain_id'),
    )


# auto set id
@event.listens_for(Event, 'before_insert')
def set_id_before_insert(mapper, connection, target):
    target.id = f"{target.hash}{target.chain_id}"


class Chain(Base):
    __tablename__ = 'chains'
    
    id = Column(BigInteger, primary_key=True)
    name = Column(String)  # not unique
    type = Column(String)  # Will be converted to enum in relationships section
    net_id = Column(BigInteger, ForeignKey('nets.id'))
    
    # Relationships will be defined after all classes
    # net = relationship(back_populates="chains")
    # datums = relationship(back_populates="chain")
    # blocks = relationship(back_populates="chain", lazy="select")
    # events = relationship(back_populates="chain", lazy="select")

    @property
    def count(self):
        if self.type == ChainTypes.esbocs:
            query = select(func.count(Block.id)).where(Block.chain_id == self.id)
        elif self.type == ChainTypes.dag_poa:
            query = select(func.count(Event.id)).where(Event.chain_id == self.id)
        else:
            raise TypeError

        return object_session(self).execute(query).scalar()

    __table_args__ = (
        UniqueConstraint('name', 'net_id', name='uq_constraint_name_and_net_id'),
    )


class Datum(Base):
    # TODO: Add counter for Atoms and Datums.
    #  Implement counter that counts datums/atoms when reading from chain
    __tablename__ = 'datums'
    
    id = Column(String, primary_key=True)  # hash + chain_id
    size = Column(Integer)
    version = Column(String)
    hash = Column(String, nullable=False)
    type = Column(String)  # Will be converted to enum in relationships section
    created_at = Column(DateTime)
    
    # Foreign keys
    chain_id = Column(BigInteger, ForeignKey('chains.id'))
    
    # Relationships will be defined after all classes
    # chain = relationship(back_populates="datums")
    # blocks = relationship(secondary=blocks_datums, back_populates="datums")
    # events = relationship(secondary=events_datums, back_populates="datums")

    sub_datum = Column(JSON)

    __table_args__ = (
        UniqueConstraint('hash', 'chain_id', name='uq_datum_hash_and_chain_id'),
        # CheckConstraint('(block_id IS NOT NULL) OR (event_id IS NOT NULL)', name='check_related_atom'),
    )


@event.listens_for(Datum, 'before_insert')
def set_id_before_insert(mapper, connection, target):
    target.id = f"{target.hash}{target.chain_id}"


class Wallet(Base):
    __tablename__ = 'wallets'
    
    id = Column(String, primary_key=True)  # addr + net_id
    addr = Column(String)  # addr - The address is unique, excluding "null" addresses
    net_id = Column(BigInteger, ForeignKey('nets.id'))
    
    # Relationships will be defined after all classes
    # net = relationship("Net", back_populates="wallets")
    # tokens = relationship(back_populates="wallet")

    __table_args__ = (
        UniqueConstraint('addr', 'net_id', name="uq_addr_net_id"),
    )


@event.listens_for(Wallet, 'before_insert')
def set_id_before_insert(mapper, connection, target):
    target.id = f"{target.addr}{target.net_id}"


class TokenBalance(Base):
    __tablename__ = 'token_balances'
    
    id = Column(BigInteger, primary_key=True)
    ticker = Column(String)
    balance = Column(String)  # datoshi
    wallet_id = Column(String, ForeignKey("wallets.id"))
    
    # Relationships will be defined after all classes
    # wallet = relationship('Wallet', back_populates="tokens")

    __table_args__ = (
        UniqueConstraint('ticker', 'wallet_id', name="un_ticker_wallet_id"),
    )


# Now define all relationships
Net.wallets = relationship("Wallet", back_populates="net")
Net.chains = relationship("Chain", back_populates="net")

Block.chain = relationship("Chain", back_populates="blocks")
Block.datums = relationship("Datum", secondary=blocks_datums, back_populates="blocks")
Block.block_recs = relationship("DatumBlockAssociation")
Block.datums_id = association_proxy("block_recs", "id", creator=lambda d_id: DatumBlockAssociation(datum_id=d_id))

Event.chain = relationship("Chain", back_populates="events")
Event.datums = relationship("Datum", secondary=events_datums, back_populates="events")
Event.event_recs = relationship("DatumEventAssociation")
Event.datums_id = association_proxy("event_recs", "id", creator=lambda d_id: DatumEventAssociation(datum_id=d_id))

Chain.net = relationship("Net", back_populates="chains")
Chain.datums = relationship("Datum", back_populates="chain")
Chain.blocks = relationship("Block", back_populates="chain", lazy="select")
Chain.events = relationship("Event", back_populates="chain", lazy="select")

Datum.chain = relationship("Chain", back_populates="datums")
Datum.blocks = relationship("Block", secondary=blocks_datums, back_populates="datums")
Datum.events = relationship("Event", secondary=events_datums, back_populates="datums")

Wallet.net = relationship("Net", back_populates="wallets")
Wallet.tokens = relationship("TokenBalance", back_populates="wallet")

TokenBalance.wallet = relationship("Wallet", back_populates="tokens")

