from . import serializers as schemas
from dataclasses import dataclass
from pydantic import BaseModel

try:
    from pycfhelpers.node.net import CFNet, CFChain
    from pycfhelpers.node.consensus import CFBlock, CFEvent
    from pycfhelpers.node.datums import CFDatum
    from pycfhelpers.node.logging import CFLog
    
    CELLFRAME_AVAILABLE = True
    
except ImportError:
    import warnings
    warnings.warn(
        "pycfhelpers not available. Shortcuts module will have limited functionality. "
        "This is expected when running outside cellframe-node environment.",
        ImportWarning
    )
    
    # Fallback - empty defaults
    CFNet = CFChain = CFBlock = CFEvent = CFDatum = CFLog = None
    CELLFRAME_AVAILABLE = False

if CELLFRAME_AVAILABLE:
    DEFAULT_SERIALIZERS = {
        CFNet: schemas.Net,
        CFChain: schemas.Chain,
        CFBlock: schemas.Block,
        CFEvent: schemas.Event,
        CFDatum: schemas.Datum,
    }
    log = CFLog()
else:
    DEFAULT_SERIALIZERS = {}
    log = None


def dump(obj, schema: BaseModel = None) -> dict:
    """
    Convert Cellframe object to dictionary using appropriate schema.
    
    Args:
        obj: Cellframe object to serialize
        schema: Optional specific schema to use
        
    Returns:
        Dictionary representation of the object
        
    Raises:
        ValueError: If no schema found and CELLFRAME_AVAILABLE is False
    """
    if not CELLFRAME_AVAILABLE:
        raise ValueError('dump() requires pycfhelpers. This function is only available in cellframe-node environment.')
        
    if schema is None:
        try:
            schema = DEFAULT_SERIALIZERS[type(obj)]
        except KeyError:
            if log:
                log.error(f"Default schema not found for object type={type(obj)}")
            raise ValueError(f'No schema found for object type: {type(obj)}')

    if hasattr(schema, 'from_orm'):
        return schema.from_orm(obj).dict()
    else:
        return schema.model_validate(obj).model_dump()
